@@ -26,6 +26,7 @@ import contextlib
 import dis
 import os
 import signal
+import subprocess
 import sys
 import threading
 import time
@@ -108,6 +109,81 @@ class Scalene:
         [threading.Thread, Union[builtins.float, None]], None
     ] = threading.Thread.join
 
+    __original_lock = threading.Lock
+
+    @staticmethod
+    def get_original_lock() -> threading.Lock:
+        return Scalene.__original_lock()
+
+    # We hijack os.system, os.popen, and subprocess.Popen so attempts to execute
+    # Python also call Scalene. TBD: integrate profiles across processes.
+    __original_os_system = os.system
+    __original_os_popen = os.popen
+    __original_subprocess_Popen = subprocess.Popen
+
+    # Likely names for the Python interpreter (assuming it's the same version as this one).
+    __alias_python_names = [
+        "python",
+        "python" + str(sys.version_info.major),
+        os.path.basename(sys.executable),
+    ]
+    __all_python_names = [
+        "python",
+        "python" + str(sys.version_info.major),
+        "python" + str(sys.version_info.major) + "." + str(sys.version_info.minor),
+        os.path.basename(sys.executable),
+    ]
+
+    @staticmethod
+    def new_os_system(cmd: str) -> Any:
+        for n in Scalene.__alias_python_names:
+            cmd = "alias " + n + "='scalene'\nexport " + n + "\n" + cmd
+        return Scalene.__original_os_system(cmd)
+
+    @staticmethod
+    def new_os_popen(cmd: str, mode: str, bufsize: int) -> Any:
+        for n in Scalene.__alias_python_names:
+            cmd = "alias " + n + "='scalene'\nexport " + n + "\n" + cmd
+        return Scalene.__original_os_popen(cmd, mode, bufsize)
+
+    @staticmethod
+    def new_subprocess_Popen(args, bufsize=-1, executable=None, stdin=None, stdout=None, stderr=None, preexec_fn=None, close_fds=True, shell=False, cwd=None, env=None, universal_newlines=None, startupinfo=None, creationflags=0, restore_signals=True, start_new_session=False, pass_fds=(), *, encoding=None, errors=None):  # type: ignore
+        if shell:
+            for n in Scalene.__alias_python_names:
+                args = "alias " + n + "='scalene'\nexport " + n + "\n" + args
+        else:
+            # Splice out the first element if it's a call to Python.
+            newargs = []
+            for a in args:
+                for n in Scalene.__all_python_names:
+                    if a == n:
+                        a = "scalene"
+                        break
+                newargs.append(a)
+            args = newargs
+        # TODO: check versions since text= was added in 3.7
+        return Scalene.__original_subprocess_Popen(
+            args,
+            bufsize,
+            executable,
+            stdin,
+            stdout,
+            stderr,
+            preexec_fn,
+            close_fds,
+            shell,
+            cwd,
+            env,
+            universal_newlines,
+            startupinfo,
+            creationflags,
+            restore_signals,
+            start_new_session,
+            pass_fds,
+            encoding=encoding,
+            errors=errors,
+        )
+
     # Statistics counters:
     #
     #   CPU samples for each location in the program
@@ -130,6 +206,11 @@ class Scalene:
         Filename, Dict[LineNumber, Dict[ByteCodeIndex, float]]
     ] = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
 
+    # mallocs attributable to Python, for each location in the program
+    __memory_python_samples: Dict[
+        Filename, Dict[LineNumber, Dict[ByteCodeIndex, float]]
+    ] = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
+
     # number of times samples were added for the above
     __memory_malloc_count: Dict[
         Filename, Dict[LineNumber, Dict[ByteCodeIndex, int]]
@@ -224,13 +305,13 @@ class Scalene:
 
     # We cache the previous signal handlers so we can play nice with
     # apps that might already have handlers for these signals.
-    __old_malloc_signal_handler: Union[
+    __original_malloc_signal_handler: Union[
         Callable[[Signals, FrameType], None], int, Handlers, None
     ] = signal.SIG_IGN
-    __old_free_signal_handler: Union[
+    __original_free_signal_handler: Union[
         Callable[[Signals, FrameType], None], int, Handlers, None
     ] = signal.SIG_IGN
-    __old_memcpy_signal_handler: Union[
+    __original_memcpy_signal_handler: Union[
         Callable[[Signals, FrameType], None], int, Handlers, None
     ] = signal.SIG_IGN
 
@@ -238,6 +319,25 @@ class Scalene:
     #   the name of the program being profiled
     __program_being_profiled = Filename("")
 
+    # Is the thread sleeping? (We use this in to properly attribute CPU time.)
+    __is_thread_sleeping: Dict[int, bool] = defaultdict(bool)  # False by default
+
+    # Threshold for highlighting lines of code in red.
+    __highlight_percentage = 33
+
+    @staticmethod
+    def is_thread_sleeping(tid: int) -> bool:
+        result = Scalene.__is_thread_sleeping[tid]
+        return result
+
+    @staticmethod
+    def set_thread_sleeping(tid: int) -> None:
+        Scalene.__is_thread_sleeping[tid] = True
+
+    @staticmethod
+    def reset_thread_sleeping(tid: int) -> None:
+        Scalene.__is_thread_sleeping[tid] = False
+
     @staticmethod
     @lru_cache(maxsize=None)
     def is_call_function(code: CodeType, bytei: ByteCodeIndex) -> bool:
@@ -256,7 +356,9 @@ periodically yields."""
         start_time = Scalene.gettime()
         interval = sys.getswitchinterval()
         while self.is_alive():
+            Scalene.__is_thread_sleeping[threading.get_ident()] = True
             Scalene.__original_thread_join(self, interval)
+            Scalene.__is_thread_sleeping[threading.get_ident()] = False
             # If a timeout was specified, check to see if it's expired.
             if timeout:
                 end_time = Scalene.gettime()
@@ -289,13 +391,13 @@ start the timer interrupts."""
         signal.signal(Scalene.__cpu_signal, Scalene.cpu_signal_handler)
         # Set signal handlers for memory allocation and memcpy events.
         # Save the previous signal handlers, if any.
-        Scalene.__old_malloc_signal_handler = signal.signal(
+        Scalene.__original_malloc_signal_handler = signal.signal(
             Scalene.__malloc_signal, Scalene.malloc_signal_handler
         )
-        Scalene.__old_free_signal_handler = signal.signal(
+        Scalene.__original_free_signal_handler = signal.signal(
             Scalene.__free_signal, Scalene.free_signal_handler
         )
-        Scalene.__old_memcpy_signal_handler = signal.signal(
+        Scalene.__original_memcpy_signal_handler = signal.signal(
             Scalene.__memcpy_signal, Scalene.memcpy_event_signal_handler
         )
         # Turn on the CPU profiling timer to run every signal_interval seconds.
@@ -317,9 +419,60 @@ process."""
             # Using wall clock time
             return time.perf_counter()
 
+    class ReplacementLock(object):
+        """Replace lock with a version that periodically yields and updates sleeping status."""
+
+        def __init__(self) -> None:
+            self.__lock: threading.Lock = Scalene.get_original_lock()
+
+        def acquire(self, blocking: bool = True, timeout: float = -1) -> bool:
+            tident = threading.get_ident()
+            if blocking == 0:
+                blocking = False
+            start_time = Scalene.gettime()
+            if blocking:
+                if timeout < 0:
+                    interval = sys.getswitchinterval()
+                else:
+                    interval = min(timeout, sys.getswitchinterval())
+            else:
+                interval = -1
+            while True:
+                Scalene.set_thread_sleeping(tident)
+                acquired_lock = self.__lock.acquire(blocking, interval)
+                Scalene.reset_thread_sleeping(tident)
+                if acquired_lock:
+                    return True
+                if not blocking:
+                    return False
+                # If a timeout was specified, check to see if it's expired.
+                if timeout != -1:
+                    end_time = Scalene.gettime()
+                    if end_time - start_time >= timeout:
+                        return False
+
+        def release(self) -> None:
+            self.__lock.release()
+
+        def locked(self) -> bool:
+            return self.__lock.locked()
+
+        def __enter__(self) -> None:
+            self.acquire()
+
+        def __exit__(self, type: str, value: str, traceback: Any) -> None:
+            self.release()
+
     def __init__(self, program_being_profiled: Optional[Filename] = None):
-        # Hijack join.
-        threading.Thread.join = Scalene.thread_join_replacement  # type: ignore
+        if True:
+            # Hijack join.
+            threading.Thread.join = Scalene.thread_join_replacement  # type: ignore
+            # Hijack lock.
+            threading.Lock = Scalene.ReplacementLock  # type: ignore
+            # Hijack system and subprocess calls.
+            os.system = Scalene.new_os_system  # type: ignore
+            os.popen = Scalene.new_os_popen  # type: ignore
+            subprocess.Popen = Scalene.new_subprocess_Popen  # type: ignore
         # Build up signal filenames (adding PID to each).
         Scalene.__malloc_signal_filename = Filename(
             Scalene.__malloc_signal_filename + str(os.getpid())
@@ -383,32 +536,47 @@ process."""
         # Update counters for every running thread.
 
         new_frames = Scalene.compute_frames_to_record(this_frame)
-
         # Now update counters (weighted) for every frame we are tracking.
         total_time = python_time + c_time
 
-        for frame in new_frames:
+        # First, find out how many frames are not sleeping.  We need
+        # to know this number so we can parcel out time appropriately
+        # (equally to each running thread).
+        total_frames = 0
+        for (frame, tident, orig_frame) in new_frames:
+            if not Scalene.__is_thread_sleeping[tident]:
+                total_frames += 1
+        if total_frames == 0:
+            return
+        normalized_time = total_time / total_frames
+
+        # Now attribute execution time.
+        for (frame, tident, orig_frame) in new_frames:
             fname = Filename(frame.f_code.co_filename)
             lineno = LineNumber(frame.f_lineno)
-            if frame == this_frame:
+            if frame == new_frames[0][0]:
                 # Main thread.
-                Scalene.__cpu_samples_python[fname][lineno] += python_time / len(
-                    new_frames
-                )
-                Scalene.__cpu_samples_c[fname][lineno] += c_time / len(new_frames)
+                if not Scalene.__is_thread_sleeping[tident]:
+                    Scalene.__cpu_samples_python[fname][lineno] += (
+                        python_time / total_frames
+                    )
+                    Scalene.__cpu_samples_c[fname][lineno] += c_time / total_frames
             else:
                 # We can't play the same game here of attributing
                 # time, because we are in a thread, and threads don't
                 # get signals in Python. Instead, we check if the
                 # bytecode instruction being executed is a function
                 # call.  If so, we attribute all the time to native.
-                normalized_time = total_time / len(new_frames)
-                if Scalene.is_call_function(frame.f_code, ByteCodeIndex(frame.f_lasti)):
-                    # Attribute time to native.
-                    Scalene.__cpu_samples_c[fname][lineno] += normalized_time
-                else:
-                    # Not in a call function so we attribute the time to Python.
-                    Scalene.__cpu_samples_python[fname][lineno] += normalized_time
+                if not Scalene.__is_thread_sleeping[tident]:
+                    # Check if the original caller is stuck inside a call.
+                    if Scalene.is_call_function(
+                        orig_frame.f_code, ByteCodeIndex(orig_frame.f_lasti)
+                    ):
+                        # It is. Attribute time to native.
+                        Scalene.__cpu_samples_c[fname][lineno] += normalized_time
+                    else:
+                        # Not in a call function so we attribute the time to Python.
+                        Scalene.__cpu_samples_python[fname][lineno] += normalized_time
 
         del new_frames
 
@@ -427,16 +595,34 @@ process."""
         Scalene.__last_signal_time = Scalene.gettime()
         Scalene.__in_signal_handler -= 1
 
+    # Returns final frame (up to a line in a file we are profiling), the thread identifier, and the original frame.
     @staticmethod
-    def compute_frames_to_record(this_frame: FrameType) -> List[FrameType]:
+    def compute_frames_to_record(
+        this_frame: FrameType,
+    ) -> List[Tuple[FrameType, int, FrameType]]:
         """Collects all stack frames that Scalene actually processes."""
-        frames: List[Optional[FrameType]] = [
-            sys._current_frames().get(cast(int, t.ident), None)
+        frames: List[Tuple[FrameType, int]] = [
+            (
+                cast(FrameType, sys._current_frames().get(cast(int, t.ident), None)),
+                cast(int, t.ident),
+            )
             for t in threading.enumerate()
+            if t != threading.main_thread()
         ]
+        # Put the main thread in the front.
+        frames.insert(
+            0,
+            (
+                sys._current_frames().get(
+                    cast(int, threading.main_thread().ident), None
+                ),
+                cast(int, threading.main_thread().ident),
+            ),
+        )
         # Process all the frames to remove ones we aren't going to track.
-        new_frames: List[FrameType] = []
-        for frame in frames:
+        new_frames: List[Tuple[FrameType, int, FrameType]] = []
+        for (frame, tident) in frames:
+            orig_frame = frame
             if not frame:
                 continue
             fname = frame.f_code.co_filename
@@ -457,7 +643,7 @@ process."""
                     if frame:
                         fname = frame.f_code.co_filename
             if frame:
-                new_frames.append(frame)
+                new_frames.append((frame, tident, orig_frame))
         return new_frames
 
     @staticmethod
@@ -493,29 +679,41 @@ process."""
             return
 
         # Process the input array.
-        arr: List[Tuple[int, str, float]] = []
+        arr: List[Tuple[int, str, float, float]] = []
         try:
             with open(Scalene.__malloc_signal_filename, "r") as mfile:
                 for count_str in mfile:
                     # for _, count_str in enumerate(mfile, 1):
                     count_str = count_str.rstrip()
-                    (action, alloc_time_str, count_str) = count_str.split(",")
-                    arr.append((int(alloc_time_str), action, float(count_str)))
+                    (
+                        action,
+                        alloc_time_str,
+                        count_str,
+                        python_fraction_str,
+                    ) = count_str.split(",")
+                    arr.append(
+                        (
+                            int(alloc_time_str),
+                            action,
+                            float(count_str),
+                            float(python_fraction_str),
+                        )
+                    )
         except FileNotFoundError:
             pass
         try:
             os.remove(Scalene.__malloc_signal_filename)
         except FileNotFoundError:
             pass
-        
+
         arr.sort()
 
         # Iterate through the array to compute the new current footprint.
         # and update the global __memory_footprint_samples.
         before = Scalene.__current_footprint
         for item in arr:
-            alloc_time, action, count = item
-            # print(fname,line_no,action, alloc_time, count)
+            alloc_time, action, count, python_fraction = item
+            # print(alloc_time, action, count, python_fraction)
             count /= 1024 * 1024
             is_malloc = action == "M"
             if is_malloc:
@@ -531,20 +729,24 @@ process."""
         # This is a pain, since we don't know to whom to attribute memory,
         # so we may overcount.
 
-        for frame in new_frames:
+        for (frame, tident, orig_frame) in new_frames:
             fname = Filename(frame.f_code.co_filename)
             line_no = LineNumber(frame.f_lineno)
             bytei = ByteCodeIndex(frame.f_lasti)
             # Add the byte index to the set for this line (if it's not there already).
             Scalene.__bytei_map[fname][line_no].add(bytei)
             curr = before
+            python_frac = 0
+            allocs = 0
             # Go through the array again and add each updated current footprint.
             for item in arr:
-                alloc_time, action, count = item
+                alloc_time, action, count, python_fraction = item
                 count /= 1024 * 1024
                 is_malloc = action == "M"
                 if is_malloc:
+                    allocs += count
                     curr += count
+                    python_frac += python_fraction * count
                 else:
                     curr -= count
                 Scalene.__per_line_footprint_samples[fname][line_no].add(curr)
@@ -555,6 +757,9 @@ process."""
             # loss per line of code.
             if after > before:
                 Scalene.__memory_malloc_samples[fname][line_no][bytei] += after - before
+                Scalene.__memory_python_samples[fname][line_no][bytei] += (
+                    python_frac / allocs
+                ) * (after - before)
                 Scalene.__memory_malloc_count[fname][line_no][bytei] += 1
                 Scalene.__total_memory_malloc_samples += after - before
             else:
@@ -595,7 +800,7 @@ process."""
 
         for item in arr:
             memcpy_time, count = item
-            for the_frame in new_frames:
+            for (the_frame, tident, orig_frame) in new_frames:
                 fname = Filename(the_frame.f_code.co_filename)
                 line_no = LineNumber(the_frame.f_lineno)
                 bytei = ByteCodeIndex(the_frame.f_lasti)
@@ -616,7 +821,12 @@ process."""
         if "site-packages" in filename or "/usr/lib/python" in filename:
             # Don't profile Python internals.
             return False
-        if "scalene.py" in filename or "adaptive.py" in filename:
+        if (
+            "scalene.py" in filename
+            or "adaptive.py" in filename
+            or "sparkline.py" in filename
+            or "scalene/__main__.py" in filename
+        ):
             # Don't profile the profiler.
             return False
         filename = os.path.abspath(filename)
@@ -695,6 +905,7 @@ process."""
         # Now, memory stats.
         # Accumulate each one from every byte index.
         n_malloc_mb = 0.0
+        n_python_malloc_mb = 0.0
         n_free_mb = 0.0
         n_avg_malloc_mb = 0.0
         n_avg_free_mb = 0.0
@@ -704,6 +915,7 @@ process."""
             # print(fname,line_no,index)
             mallocs = Scalene.__memory_malloc_samples[fname][line_no][index]
             n_malloc_mb += mallocs
+            n_python_malloc_mb += Scalene.__memory_python_samples[fname][line_no][index]
             n_malloc_count += Scalene.__memory_malloc_count[fname][line_no][index]
             frees = Scalene.__memory_free_samples[fname][line_no][index]
             n_free_mb += frees
@@ -723,13 +935,14 @@ process."""
             if not Scalene.__total_memory_malloc_samples
             else n_malloc_mb / Scalene.__total_memory_malloc_samples
         )
-
+        n_python_fraction = 0 if not n_malloc_mb else n_python_malloc_mb / n_malloc_mb
+        # print(n_python_malloc_mb, n_malloc_mb)
         # Finally, print results.
         n_cpu_percent_c_str: str = (
-            "" if not n_cpu_percent_c else "%6.2f%%" % n_cpu_percent_c
+            "" if not n_cpu_percent_c else "%6.1f%%" % n_cpu_percent_c
         )
         n_cpu_percent_python_str: str = (
-            "" if not n_cpu_percent_python else "%6.2f%%" % n_cpu_percent_python
+            "" if not n_cpu_percent_python else "%6.1f%%" % n_cpu_percent_python
         )
         n_growth_mb_str: str = (
             "" if (not n_growth_mb and not n_usage_fraction) else "%5.0f" % n_growth_mb
@@ -737,9 +950,12 @@ process."""
         n_usage_fraction_str: str = (
             "" if not n_usage_fraction else "%3.0f%%" % (100 * n_usage_fraction)
         )
+        n_python_fraction_str: str = (
+            "" if not n_python_fraction else "%3.0f%%" % (100 * n_python_fraction)
+        )
         n_copy_b = Scalene.__memcpy_samples[fname][line_no]
         n_copy_mb_s = n_copy_b / (1024 * 1024 * Scalene.__elapsed_time)
-        n_copy_mb_s_str: str = "" if n_copy_mb_s < 0.1 else "%6.0f" % n_copy_mb_s
+        n_copy_mb_s_str: str = "" if n_copy_mb_s < 0.5 else "%6.0f" % n_copy_mb_s
 
         if did_sample_memory:
             spark_str: str = ""
@@ -751,12 +967,22 @@ process."""
                 _, _, spark_str = Scalene.generate_sparkline(
                     samples.get()[0 : samples.len()], 0, current_max
                 )
+
+            # Red highlight
+            if (
+                n_usage_fraction >= Scalene.__highlight_percentage
+                or (n_cpu_percent_c + n_cpu_percent_python)
+                >= Scalene.__highlight_percentage
+            ):
+                print(u"\u001b[31m", end="")
+
             print(
-                "%6d |%9s |%9s | %5s | %-9s %-4s |%-6s | %s"
+                "%6d |%7s |%7s | %5s | %5s | %-9s %-4s |%-6s | %s"
                 % (
                     line_no,
                     n_cpu_percent_python_str,
                     n_cpu_percent_c_str,
+                    n_python_fraction_str,
                     n_growth_mb_str,
                     spark_str,
                     n_usage_fraction_str,
@@ -766,12 +992,22 @@ process."""
                 file=out,
             )
         else:
+
+            # Red highlight
+            if (
+                n_cpu_percent_c + n_cpu_percent_python
+            ) >= Scalene.__highlight_percentage:
+                print(u"\u001b[31m", end="")
+
             print(
-                "%6d |%9s |%9s | %s"
+                "%6d |%7s |%7s | %s"
                 % (line_no, n_cpu_percent_python_str, n_cpu_percent_c_str, line),
                 file=out,
             )
 
+        # Reset color
+        print(u"\u001b[0m", end="")
+
     @staticmethod
     def output_profiles() -> bool:
         """Write the profile out (currently to stdout)."""
@@ -840,10 +1076,11 @@ process."""
                 )
 
                 print(
-                    "       |%9s |%9s | %s %s %s"
+                    "       |%7s |%7s | %s %s %s %s"
                     % (
                         "CPU %",
                         "CPU %",
+                        "Mem % |" if did_sample_memory else "",
                         " Net  |" if did_sample_memory else "",
                         "Memory usage   |" if did_sample_memory else "",
                         "Copy  |" if did_sample_memory else "",
@@ -851,10 +1088,11 @@ process."""
                     file=out,
                 )
                 print(
-                    "  Line |%9s |%9s | %s %s %s [%s]"
+                    "  Line |%7s |%7s | %s %s %s %s [%s]"
                     % (
-                        "(Python)",
-                        "(native)",
+                        "Python",
+                        "native",
+                        "Python|" if did_sample_memory else "",
                         " (MB) |" if did_sample_memory else "",
                         "over time /  % |" if did_sample_memory else "",
                         "(MB/s)|" if did_sample_memory else "",
@@ -862,7 +1100,7 @@ process."""
                     ),
                     file=out,
                 )
-                print("-" * 80, file=out)
+                print("-" * 119, file=out)
 
                 with open(fname, "r") as source_file:
                     for line_no, line in enumerate(source_file, 1):
@@ -876,9 +1114,9 @@ process."""
     def disable_signals() -> None:
         """Turn off the profiling signals."""
         signal.setitimer(Scalene.__cpu_timer_signal, 0)
-        signal.signal(Scalene.__malloc_signal, Scalene.__old_malloc_signal_handler)
-        signal.signal(Scalene.__free_signal, Scalene.__old_free_signal_handler)
-        signal.signal(Scalene.__memcpy_signal, Scalene.__old_memcpy_signal_handler)
+        signal.signal(Scalene.__malloc_signal, Scalene.__original_malloc_signal_handler)
+        signal.signal(Scalene.__free_signal, Scalene.__original_free_signal_handler)
+        signal.signal(Scalene.__memcpy_signal, Scalene.__original_memcpy_signal_handler)
 
     @staticmethod
     def exit_handler() -> None:
@@ -975,7 +1213,7 @@ process."""
                         exec(code, the_globals, the_locals)
                     except BaseException:  # as be
                         # Intercept sys.exit.
-                        # print(traceback.format_exc())
+                        # print(traceback.format_exc())  # FIXME
                         pass
                     profiler.stop()
                     # If we've collected any samples, dump them.
